#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


def _exec(code, qs):
    """
    Executes the Python code in 'filename'.

    Args:
        code (string): ProjectQ code.
        qs (tuple<Qureg>): Qubits to which the permutation is being
                           applied.
    """
    from projectq.ops import C, X, Z, All
    exec(code)
